/*
 * HTMLPlugin.java - HTML Plugin
 * Copyright (C) 1999,2000 Romain Guy, Andr Kaplan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


import gnu.regexp.REException;
import gnu.regexp.RE;

import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;

import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EBPlugin;
import org.gjt.sp.jedit.EditPane;
import org.gjt.sp.jedit.GUIUtilities;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.MiscUtilities;
import org.gjt.sp.jedit.OptionGroup;
import org.gjt.sp.jedit.gui.HistoryModel;
import org.gjt.sp.jedit.gui.OptionsDialog;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.EditPaneUpdate;
import org.gjt.sp.jedit.msg.PropertiesChanged;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.jedit.textarea.TextAreaPainter;

import org.gjt.sp.util.Log;


public class HTMLPlugin extends EBPlugin
{
    private static boolean showTagHighlightDefault;
    private static boolean useTagHighlightFilter;
    private static RE      tagHighlightFilter;

    private static HTMLAbstractOptionPane[] optionPanes = null;
    private static OptionPaneComponent[][] optionPaneComponents = null;

    static {
        HTMLPlugin.propertiesChanged();

        optionPanes = new HTMLAbstractOptionPane[] {
              new HTMLGeneralOptionPane()
            , new HTMLWrapIndentOptionPane()
            , new HTMLFormatOptionPane()
            , new HTMLCleanOptionPane()
            , new HTMLXMLOptionPane()
            , new HTMLNewTagsOptionPane()
        };

        optionPaneComponents = new OptionPaneComponent[optionPanes.length][];

        for (int i = 0; i < optionPanes.length; i++) {
            optionPaneComponents[i] = optionPanes[i].getOptionPaneComponents();
        }
    }


    public void start() {
        jEdit.addAction(new jtidy_tidy_buffer());
        jEdit.addAction(new jtidy_generate_config_file());
        jEdit.addAction(new next_tag());
        jEdit.addAction(new prev_tag());
        jEdit.addAction(new remove_tags());
        jEdit.addAction(new match_tag());
        jEdit.addAction(new close_last_open_element());
        jEdit.addAction(new entities_to_chars(false));
        jEdit.addAction(new entities_to_chars(true));
        jEdit.addAction(new chars_to_entities(false));
        jEdit.addAction(new chars_to_entities(true));
        jEdit.addAction(new color_picker());

        JTidyErrorSourceWriter.start();
    }


    public void stop() {
        JTidyErrorSourceWriter.stop();
    }


    public void createMenuItems(Vector menuItems) {
        menuItems.addElement(GUIUtilities.loadMenu("HTML-menu"));
    }


    public void createOptionPanes(OptionsDialog optionsDialog) {
        OptionGroup htmlGroup = new OptionGroup("HTML");

        htmlGroup.addOptionPane(new HTMLOptionPane());

        OptionGroup tidyGroup = new OptionGroup("HTML-Tidy");

        for (int i = 0; i < optionPanes.length; i++) {
            tidyGroup.addOptionPane(optionPanes[i]);
        }

        htmlGroup.addOptionGroup(tidyGroup);

        optionsDialog.addOptionGroup(htmlGroup);
    }


    public void handleMessage(EBMessage message) {
        if (message instanceof EditPaneUpdate) {
            EditPaneUpdate epu = (EditPaneUpdate) message;
            EditPane editPane = epu.getEditPane();

            if (epu.getWhat() == EditPaneUpdate.CREATED) {
                JEditTextArea textArea = editPane.getTextArea();
                TextAreaPainter textAreaPainter = textArea.getPainter();

                TagHighlight tagHighlight =
                    (TagHighlight) TagHighlight.addHighlightTo(editPane);

                textAreaPainter.addCustomHighlight(tagHighlight);
            } else if (epu.getWhat() == EditPaneUpdate.DESTROYED) {
                TagHighlight.removeHighlightFrom(editPane);
            } else if (epu.getWhat() == EditPaneUpdate.BUFFER_CHANGED) {
                TagHighlight.bufferChanged(editPane);
            }
        } else if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate) message;
            Buffer buffer   = bu.getBuffer();

            if (bu.getWhat() == BufferUpdate.CREATED) {
                TagHighlight.bufferCreated(buffer,
                    HTMLPlugin.getTagHighlightDefaultFor(buffer)
                );
            } else if (bu.getWhat() == BufferUpdate.CLOSED) {
                TagHighlight.bufferClosed(buffer);
            }
        } else if (message instanceof PropertiesChanged) {
            HTMLPlugin.propertiesChanged();
            TagHighlight.propertiesChanged();
        }
    }


    public static boolean getTagHighlightDefaultFor(Buffer buffer) {
        boolean enabled = false;

        if (showTagHighlightDefault) {
            if (useTagHighlightFilter) {
                if (   tagHighlightFilter != null
                    && tagHighlightFilter.isMatch(buffer.getName())
                ) {
                    Log.log(Log.DEBUG, HTMLPlugin.class, "**** [" + buffer.getName() + "]");
                    enabled = true;
                }
            } else {
                enabled = true;
            }
        }

        return enabled;
    }


    private static void propertiesChanged() {
        showTagHighlightDefault = jEdit.getBooleanProperty(
                    "html.show-tag-highlight-default", true
                );

        useTagHighlightFilter = jEdit.getBooleanProperty(
            "html.use-tag-highlight-filter", true
        );

        HistoryModel filterHistory = HistoryModel.getModel("html.tag-highlight-filter");
        String glob;
        if  (filterHistory.getSize() > 0) {
            glob = filterHistory.getItem(0);
        } else {
            glob = jEdit.getProperty("html.tag-highlight-default-filter", "*.{html,htm}");
        }
        tagHighlightFilter = null;
        try {
            tagHighlightFilter = new RE(MiscUtilities.globToRE(glob));
        } catch (REException e) {
            Log.log(Log.ERROR, HTMLPlugin.class, e);
        }
    }


    public static void saveTo(Properties props) {
        if (optionPaneComponents == null) { return; }

        for (int i = 0; i < optionPaneComponents.length; i++) {
            for (int j = 0; j < optionPaneComponents[i].length; j++) {
                optionPaneComponents[i][j].saveTo(props);
            }
        }
    }
}

