/*
 * jtidy_tidy_buffer.java
 * Copyright (c) 2000 Andre Kaplan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


import java.io.*;
import java.util.Properties;

import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.EBComponent;
import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EditAction;
import org.gjt.sp.jedit.EditBus;
import org.gjt.sp.jedit.EditPane;
import org.gjt.sp.jedit.GUIUtilities;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.Mode;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.EditPaneUpdate;
import org.gjt.sp.jedit.textarea.JEditTextArea;

import org.gjt.sp.util.Log;

import org.w3c.tidy.Node;
import org.w3c.tidy.Tidy;


public class jtidy_tidy_buffer
    extends EditAction
    implements EBComponent
{
    public jtidy_tidy_buffer() {
        super("jtidy-tidy-buffer");
    }


    public void invoke(View view)
    {
        Tidy jtidy = new Tidy();

        // The following comment is for JDK1.2+
        // Properties props = new Properties();
        // HTMLPlugin.saveTo(props);
        // jtidy.setConfigurationFromProps(props);

        File jtidyTempDir = new File(jEdit.getSettingsDirectory(),
            "jtidy" + File.separator + "tmp"
        );
        if (!jtidyTempDir.exists()) {
            jtidyTempDir.mkdirs();
        }

        File jtidyConfigFile =  new File(jtidyTempDir.getAbsolutePath()
            + File.separator
            + "jtidy_config_" + System.currentTimeMillis() + ".txt");

        // Saving current config to jtidyConfigFile
        try {
            Properties props = new Properties();
            HTMLPlugin.saveTo(props);

            OutputStream jtidyConfigOut = new BufferedOutputStream(
                new FileOutputStream(jtidyConfigFile)
            );

            props.save(jtidyConfigOut, "Generated Tidy configuration file");
            // JDK1.2+
            // props.store(jtidyConfigOut, "Generated Tidy configuration file");

            jtidyConfigOut.flush();
            jtidyConfigOut.close();

            if (jtidyConfigFile.exists()) {
                jtidy.setConfigurationFromFile(jtidyConfigFile.getAbsolutePath());
                jtidyConfigFile.delete();
            }
        } catch (IOException ioe) {
            Log.log(Log.ERROR, this, ioe);
            return;
        }

        Buffer buffer = view.getBuffer();
        Mode mode     = buffer.getMode();

        // Setting tidy input
        JEditTextArea textArea = view.getTextArea();
        String text = textArea.getText();
        InputStream in = new BufferedInputStream(
            new ByteArrayInputStream(text.getBytes())
        );

        // Setting tidy standard output
        boolean noOutput = jEdit.getBooleanProperty("html.no-output", false);
        boolean writeBack = jEdit.getBooleanProperty("html.write-back", false);

        ByteArrayOutputStream baOut = null;
        OutputStream out = null;
        if (!noOutput && !writeBack) {
            baOut = new ByteArrayOutputStream();
            out = new NewlineOutputFilter(new BufferedOutputStream(baOut));
        }

        // Setting tidy error output
        StringWriter sw = new StringWriter();
        JTidyErrorSourceWriter esw = new JTidyErrorSourceWriter(buffer, sw);
        PrintWriter jtidyErr = new PrintWriter(esw);
        jtidy.setErrout(jtidyErr);

        // Tidy buffer
        Node node = jtidy.parse(in, out);
        jtidyErr.close();

        // Tidy logo, errors, advice to activity log
        String notice = sw.toString();
        Log.log(Log.NOTICE, this, notice);

        if (node == null || esw.getErrorCount() > 0) {
            GUIUtilities.error(view, "jtidy.msg.fatal-error",
                new Object[] { buffer.getName() }
            );
            return;
        }

        if (noOutput || writeBack) {
            // No output. write back simply return since
            // no new buffer is created
            return;
        }

        // Create a new buffer and put the tidied content
        EditBus.addToBus(this);

        EditPane newEditPane = view.getEditPane();
        this.job = new FillBuffer(newEditPane, mode, baOut.toString());

        Buffer newBuffer = jEdit.newFile(view);
    }


    public void handleMessage(EBMessage message) {
        if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate) message;
            if (bu.getWhat() == BufferUpdate.CREATED) {
                if (this.job != null) {
                    this.job.setBuffer(bu.getBuffer());
                }
            }
        } else if (message instanceof EditPaneUpdate) {
            EditPaneUpdate epu = (EditPaneUpdate) message;
            if (epu.getWhat() == EditPaneUpdate.BUFFER_CHANGED) {
                if (this.job != null) {
                    this.job.run();
                    this.job = null;
                }
                EditBus.removeFromBus(this);
            }
        }
    }


    private FillBuffer job;


    private static class FillBuffer
    {
        private Buffer   buffer;
        private EditPane editPane;
        private Mode     mode;
        private String   text;


        private FillBuffer() {}


        public FillBuffer(EditPane editPane, Mode mode, String text) {
            this.editPane = editPane;
            this.mode     = mode;
            this.text     = text;
        }


        public void run() {
            if (this.buffer != this.editPane.getBuffer()) {
                Log.log(Log.DEBUG, this, "buffer != editPane.getBuffer()");
            } else {
                this.editPane.getTextArea().setText(this.text);
                this.buffer.setMode(this.mode);
            }
        }


        public void setBuffer(Buffer buffer) {
            this.buffer = buffer;
        }
    }
}
